"""
This module defines an UFL transformation, that takes a UFL expression
and transforms it into a sum of terms that all contain the correct number
of test function terms (which is equal to the form rank).
"""
from __future__ import absolute_import

from dune.perftool.ufl.modified_terminals import extract_modified_arguments
from dune.perftool.ufl.transformations import ufl_transformation
from dune.perftool.ufl.transformations.replace import replace_expression

from ufl.algorithms import MultiFunction
from ufl.classes import Zero

import itertools


class _ReplacementDict(dict):
    def __init__(self, good=[], bad=[]):
        dict.__init__(self)
        for a in bad:
            self[a] = Zero(shape=a.ufl_shape, free_indices=a.ufl_free_indices, index_dimensions=a.ufl_index_dimensions)
        for a in good:
            self[a] = a


@ufl_transformation(name="accterms2", extraction_lambda=lambda l: l)
def split_into_accumulation_terms(expr):
    mod_args = extract_modified_arguments(expr)

    accumulation_terms = []

    # Treat the case of a rank 1 form:
    if len(list(filter(lambda ma: ma.argexpr.number() == 1, mod_args))) == 0:
        for arg in mod_args:
            # Do the replacement on the expression
            accumulation_terms.append(replace_expression(expr,
                                                         replacemap=_ReplacementDict(good=(arg.expr,),
                                                                                     bad=[ma.expr for ma in mod_args],
                                                                                     )
                                                         )
                                      )
    # and now the case of a rank 2 form:
    else:
        for arg1, arg2 in itertools.product(filter(lambda ma: ma.argexpr.number() == 0, mod_args),
                                            filter(lambda ma: ma.argexpr.number() == 1, mod_args)
                                            ):
            accumulation_terms.append(replace_expression(expr,
                                                         replacemap=_ReplacementDict(good=(arg1.expr, arg2.expr),
                                                                                     bad=[ma.expr for ma in mod_args],
                                                                                     )
                                                         )
                                      )

    # and return the result
    return accumulation_terms
